untyped
global function SpeechBubblesInit


struct playerMessageData
{
	string playerName
	string message
	var RUI
	bool isRUIassigned
	float timeSent
}

playerMessageData[32] playerMessages //there shouldn't be more players present

//const float defaultMessageDuration = 120 //0.03 is the vertical offset for worldpos, 12 is the time

const float maxAlphaDistance = 5 // 196,954 is about 5 meters in hammer, 39,3907h = 1m
//const float zeroAlphaDistance = 35 // 1969,54 is about 50 meters in hammer
//const float fontSize = 64 //32 is default with 13 pixels per letter

//0.135 vertical offset on 64 font size, 0.095 on 32 font size
//27 pixels on 64 font, 13.6 pixels on 32 font, perfectly linear
//y = ax + b
//a = 0.00125
//b = 0.055
//where x is the font size, and y is the vertical offset

void function SpeechBubblesInit()
{	
	initializeTableData()
	AddCallback_OnPlayerDisconnected(ClearPlayersTableData)
	AddCallback_OnReceivedSayTextMessage(TextMessageHandler)
	thread SpeechBubblesHandler()
}


void function SpeechBubblesHandler()
{
	vector temporaryMessagePosition
	float temporaryDistance
	float temporaryAlpha
	while(true)
	{
		Wait(0.01)
		array<entity> players = GetPlayerArray()
		foreach(player in players)
		{	
			if(player != GetLocalClientPlayer())
			{
				if(PlayerCanSee(GetLocalClientPlayer(), player, true, 90.0)) //if a player is visible
				{
					for(int i = 0; i<32; i++)
					{
						if(playerMessages[i].playerName == player.GetPlayerName()) //when encountering that player's name in the table
						{
							if(Time() - playerMessages[i].timeSent < GetConVarFloat("speechbubbles_message_duration")) //Check time sent
							{
								//temporaryMessagePosition = WorldToScreenPos(player.EyePosition()) //GetWorldSpaceCenter() or EyePosition(), EyePositon tends to be wobbly
								temporaryMessagePosition = WorldToScreenPos(player.GetOrigin() + <0,0,player.GetBoundingMaxs().z>)
								//temporaryMessagePosition.x -= playerMessages[i].message.len()*13/(2*GetScreenSize()[0])
								//temporaryMessagePosition.y -= GetConVarFloat("speechbubbles_offset_static") //oofset should be 0.095, also can be set with a convar GetConVarFloat("speechbubbles_offset_static")
								temporaryMessagePosition.x -= playerMessages[i].message.len()*(GetConVarFloat("speechbubbles_message_font_size")/2.37)/(2*GetScreenSize()[0]) //2.37 is a magical const for font sizes
								temporaryMessagePosition.y -= ((GetConVarFloat("speechbubbles_message_font_size")*0.00125) + 0.055)
								temporaryDistance = GetDistance(GetLocalClientPlayer(), player)
								temporaryAlpha = CalculateAlphaFromDistance(temporaryDistance)
								//temporaryMessagePosition.y -= (1/(temporaryDistance*GetConVarFloat("speechbubbles_offset_dynamic")) + GetConVarFloat("speechbubbles_offset_static")) //used to be 4*13/(2*GetScreenSize()[0])
								if(playerMessages[i].RUI != null) //check if rui exists
								{
									updateSpeechBubbleRUI(temporaryMessagePosition, playerMessages[i], i, temporaryAlpha)
								}
								else
								{
									playerMessages[i].RUI = createSpeechBubbleRUI(temporaryMessagePosition, playerMessages[i], temporaryAlpha)
								}
							}
							else
							{
								if(playerMessages[i].RUI != null)
								{
									RuiDestroy(playerMessages[i].RUI)
									playerMessages[i].RUI = null
								}
							}
							break
						}
					}
				}
				else //player is not currently seen
				{
					for(int i = 0; i<32; i++)
					{
						if(playerMessages[i].playerName == player.GetPlayerName()) //when encountering that player's name in the table
						{
							//check for rui, if exists, destroy
							if(playerMessages[i].RUI != null)
							{
								RuiDestroy(playerMessages[i].RUI)
								playerMessages[i].RUI = null
							}
							break
						}
					}
				}
			}
		}
	}
}


//need to initialize the struct with empty names, and delete existing ones on disconnect
ClClient_MessageStruct function TextMessageHandler(ClClient_MessageStruct message)
{
	//message.message, message.player
	//make a check so my own aren't displayed
	if(message.player != GetLocalClientPlayer())
	{
		for(int i = 0; i<32; i++)
		{
			if(playerMessages[i].playerName == message.player.GetPlayerName() || playerMessages[i].playerName == "playerMessagesStructEmptyString") //when encountering that player's name in the table
			{
				playerMessages[i].playerName = message.player.GetPlayerName()
				playerMessages[i].message = message.message
				playerMessages[i].timeSent = Time()
				
				break
			}
		}
	}
	return message
}


vector function WorldToScreenPos( vector position )
{
    array pos = expect array( Hud.ToScreenSpace( position ) )

    vector result = <float( pos[0] ), float( pos[1] ), 0 >
    //Logger.Info(result.x.tostring() + " " result.y.tostring())
	vector finalRUIposition = <result.x/GetScreenSize()[0], result.y/GetScreenSize()[1], 0>
	//Logger.Info(GetScreenSize()[0].tostring() + "/" + GetScreenSize()[1].tostring())
    return finalRUIposition
}


float function GetDistance(entity client, entity player)
{
	TraceResults trace = TraceLine( client.EyePosition(), player.GetWorldSpaceCenter(), null, TRACE_MASK_BLOCKLOS, TRACE_COLLISION_GROUP_NONE )
	float hDist = Distance(client.EyePosition(), trace.endPos )
	if(hDist == 0)
	{
		return 0.1
	}
	float dDist = hDist * 0.01904 * (4/3)
	return dDist
}


float function CalculateAlphaFromDistance(float distance)
{
	if(distance > maxAlphaDistance)
	{
		if(distance < GetConVarFloat("speechbubbles_message_max_distance"))
		{
			return (1-(distance/(GetConVarFloat("speechbubbles_message_max_distance")-maxAlphaDistance)))*10
		}
		return 0
	}
	return 10
}


var function createSpeechBubbleRUI(vector position, playerMessageData data, float alpha)
{
	var rui = RuiCreate( $"ui/cockpit_console_text_top_left.rpak", clGlobal.topoCockpitHudPermanent, RUI_DRAW_COCKPIT, -1 )
	RuiSetInt(rui, "maxLines", 1)
	RuiSetInt(rui, "lineNum", 1)
	RuiSetFloat2(rui, "msgPos", position)
	RuiSetString(rui, "msgText", data.message)
	RuiSetFloat(rui, "msgFontSize", GetConVarFloat("speechbubbles_message_font_size"))
	RuiSetFloat(rui, "msgAlpha", alpha)
	RuiSetFloat(rui, "thicken", 0.0)
	RuiSetFloat3(rui, "msgColor", <1,1,1>)
	return rui
}


void function updateSpeechBubbleRUI(vector position, playerMessageData data, int index, float alpha)
{
	RuiSetFloat2(playerMessages[index].RUI, "msgPos", position)
	RuiSetString(playerMessages[index].RUI, "msgText", data.message)
	RuiSetFloat(playerMessages[index].RUI, "msgAlpha", alpha)
}


void function ClearPlayersTableData(entity player)
{
	string name = player.GetPlayerName()
	//i need to move all names, messages and times up so that there's freed up space at the end, as to not lose space
	for(int i = 0; i<32; i++)
	{
		if(playerMessages[i].playerName == name)
		{
			int temp = i
			
			while(temp<31)
			{
				playerMessages[temp].playerName = playerMessages[temp+1].playerName
				playerMessages[temp].message = playerMessages[temp+1].message
				playerMessages[temp].timeSent = playerMessages[temp+1].timeSent
				playerMessages[temp].RUI = playerMessages[temp+1].RUI
				temp++
			}
			playerMessages[31].playerName = "playerMessagesStructEmptyString"
			playerMessages[31].message = ""
			playerMessages[31].timeSent = Time() - GetConVarFloat("speechbubbles_message_duration")
			playerMessages[31].RUI = null
		}//no break here in case a player is saved twice
	}
}


void function initializeTableData()
{
	for(int i = 0; i<32; i++)
	{

		playerMessages[i].playerName = "playerMessagesStructEmptyString"
		playerMessages[i].message = ""
		playerMessages[i].timeSent = Time() - GetConVarFloat("speechbubbles_message_duration")
		playerMessages[i].RUI = null
			
	}
}


//so, we have a bunch of RUIS
//each RUI is tied to a player
//destroy it when player leaves vision
//create it when player has present message and is in vision
//Player sends message
//message added to thread with player's name and time
//in thread, if you can see player, do:
//check their status in the table
//if there's a present message, check for time and then check if an RUI is already present
//if it is, set the message and location or remove if time is up
//if it isn't, create it




//add callback for player disconnect to remove their RUI

//issues: wobbly when using EyePos, offsets can be difficult
//messages show up when not seeing an enemy, might be because of respawns
//they disappear shortly after